import 'package:flutter/material.dart';
import '../render/forme_render_data.dart';
import '../render/forme_render_utils.dart';

import '../forme_state_model.dart';
import 'forme_decoration.dart';
import '../forme_field.dart';
import '../forme_core.dart';

class FormeListTileItem<T> {
  final Widget title;
  final bool readOnly;
  final bool visible;
  final EdgeInsets padding;

  /// only work when split = 1
  final Widget? secondary;
  final ListTileControlAffinity controlAffinity;

  /// only work when split = 1
  final Widget? subtitle;
  final bool dense;
  final T data;
  final bool ignoreSplit;

  FormeListTileItem(
      {required this.title,
      this.subtitle,
      this.secondary,
      ListTileControlAffinity? controlAffinity,
      this.readOnly = false,
      this.visible = true,
      this.dense = false,
      EdgeInsets? padding,
      required this.data,
      this.ignoreSplit = false})
      : this.controlAffinity =
            controlAffinity ?? ListTileControlAffinity.leading,
        this.padding = padding ?? EdgeInsets.zero;
}

enum FormeListTileType { Checkbox, Switch, Radio }

class FormeListTile<T>
    extends NonnullValueField<List<T>, FormeListTileModel<T>> {
  FormeListTile({
    ValueChanged<List<T>>? onChanged,
    NonnullFieldValidator<List<T>>? validator,
    AutovalidateMode? autovalidateMode,
    List<T>? initialValue,
    NonnullFormFieldSetter<List<T>>? onSaved,
    String? name,
    bool readOnly = false,
    FormeListTileType type = FormeListTileType.Checkbox,
    required List<FormeListTileItem<T>>? items,
    FormeListTileModel<T>? model,
    Key? key,
  }) : super(
            key: key,
            model: (model ?? FormeListTileModel<T>()).copyWith(items: items),
            readOnly: readOnly,
            name: name,
            onChanged: onChanged,
            onSaved: onSaved,
            autovalidateMode: autovalidateMode,
            initialValue: initialValue ?? [],
            validator: validator,
            builder: (state) {
              bool readOnly = state.readOnly;
              int split = state.model.split ?? 2;
              List<FormeListTileItem<T>> items = state.model.items ?? [];
              bool allowSelectAll = state.model.allowSelectAll ?? true;

              FormeListTileRenderData? formeListTileRenderData =
                  state.model.formeListTileRenderData;
              FormeCheckboxRenderData? formeCheckboxRenderData =
                  state.model.formeCheckboxRenderData;
              FormeRadioRenderData? formeRadioRenderData =
                  state.model.formeRadioRenderData;
              FormeSwitchRenderData? formeSwitchRenderData =
                  state.model.formeSwitchRenderData;

              List<Widget> wrapWidgets = [];

              void changeValue(T value) {
                state.requestFocus();
                switch (type) {
                  case FormeListTileType.Checkbox:
                  case FormeListTileType.Switch:
                    List<T> values = List.of(state.value);
                    if (!values.remove(value)) {
                      values.add(value);
                    }
                    state.didChange(values);
                    break;
                  case FormeListTileType.Radio:
                    state.didChange([value]);
                    break;
                }
              }

              Widget createFormeListTileItem(
                  FormeListTileItem item, bool selected, bool readOnly) {
                switch (type) {
                  case FormeListTileType.Radio:
                    return RadioListTile<T>(
                      shape: formeRadioRenderData?.shape,
                      tileColor: formeRadioRenderData?.tileColor,
                      selectedTileColor:
                          formeRadioRenderData?.selectedTileColor,
                      activeColor: formeRadioRenderData?.activeColor,
                      secondary: item.secondary,
                      subtitle: item.subtitle,
                      groupValue: state.value.isEmpty ? null : state.value[0],
                      controlAffinity: item.controlAffinity,
                      contentPadding: item.padding,
                      dense: item.dense,
                      title: item.title,
                      value: item.data,
                      onChanged:
                          readOnly ? null : (v) => changeValue(item.data),
                    );
                  case FormeListTileType.Checkbox:
                    return CheckboxListTile(
                      shape: formeCheckboxRenderData?.shape,
                      tileColor: formeCheckboxRenderData?.tileColor,
                      selectedTileColor:
                          formeCheckboxRenderData?.selectedTileColor,
                      activeColor: formeCheckboxRenderData?.activeColor,
                      checkColor: formeCheckboxRenderData?.checkColor,
                      secondary: item.secondary,
                      subtitle: item.subtitle,
                      controlAffinity: item.controlAffinity,
                      contentPadding: item.padding,
                      dense: item.dense,
                      title: item.title,
                      value: selected,
                      onChanged:
                          readOnly ? null : (v) => changeValue(item.data),
                    );
                  case FormeListTileType.Switch:
                    return SwitchListTile(
                      tileColor: formeSwitchRenderData?.tileColor,
                      activeColor: formeSwitchRenderData?.activeColor,
                      activeTrackColor: formeSwitchRenderData?.activeTrackColor,
                      inactiveThumbColor:
                          formeSwitchRenderData?.inactiveThumbColor,
                      inactiveTrackColor:
                          formeSwitchRenderData?.inactiveTrackColor,
                      activeThumbImage: formeSwitchRenderData?.activeThumbImage,
                      inactiveThumbImage:
                          formeSwitchRenderData?.inactiveThumbImage,
                      shape: formeSwitchRenderData?.shape,
                      selectedTileColor:
                          formeSwitchRenderData?.selectedTileColor,
                      secondary: item.secondary,
                      subtitle: item.subtitle,
                      controlAffinity: item.controlAffinity,
                      contentPadding: item.padding,
                      dense: item.dense,
                      title: item.title,
                      value: selected,
                      onChanged:
                          readOnly ? null : (v) => changeValue(item.data),
                    );
                }
              }

              Widget createCommonItem(
                  FormeListTileItem item, bool selected, bool readOnly) {
                switch (type) {
                  case FormeListTileType.Checkbox:
                    return FormeRenderUtils.checkbox(
                        selected,
                        readOnly || item.readOnly
                            ? null
                            : (v) => changeValue(item.data),
                        formeCheckboxRenderData);
                  case FormeListTileType.Switch:
                    return FormeRenderUtils.adaptiveSwitch(
                        selected,
                        readOnly || item.readOnly
                            ? null
                            : (v) => changeValue(item.data),
                        formeSwitchRenderData);
                  case FormeListTileType.Radio:
                    return FormeRenderUtils.radio<T>(
                        item.data,
                        state.value.isEmpty ? null : state.value[0],
                        readOnly || item.readOnly
                            ? null
                            : (v) => changeValue(item.data),
                        formeRadioRenderData);
                }
              }

              for (int i = 0; i < items.length; i++) {
                FormeListTileItem<T> item = items[i];
                bool isReadOnly = readOnly || item.readOnly;
                bool selected = state.value.contains(item.data);
                if (split > 0) {
                  double factor = 1 / split;
                  if (factor == 1) {
                    wrapWidgets.add(
                        createFormeListTileItem(item, selected, isReadOnly));
                    continue;
                  }
                }

                Widget tileItem = createCommonItem(item, selected, readOnly);

                final Widget title = split == 0
                    ? item.title
                    : Flexible(
                        child: item.title,
                      );

                List<Widget> children;
                switch (item.controlAffinity) {
                  case ListTileControlAffinity.leading:
                    children = [tileItem, title];
                    break;
                  default:
                    children = [title, tileItem];
                    break;
                }

                Row tileItemRow = Row(
                  mainAxisSize: MainAxisSize.min,
                  children: children,
                );

                Widget groupItemWidget = Padding(
                  padding: item.padding,
                  child: InkWell(
                      borderRadius:
                          const BorderRadius.all(Radius.circular(4.0)),
                      onTap: isReadOnly
                          ? null
                          : () {
                              changeValue(item.data);
                            },
                      child: tileItemRow),
                );

                bool visible = item.visible;
                if (split <= 0) {
                  wrapWidgets.add(Visibility(
                    child: groupItemWidget,
                    visible: visible,
                  ));
                  if (visible && i < items.length - 1)
                    wrapWidgets.add(SizedBox(
                      width: 8.0,
                    ));
                } else {
                  double factor = item.ignoreSplit ? 1 : 1 / split;
                  wrapWidgets.add(Visibility(
                    child: FractionallySizedBox(
                      widthFactor: factor,
                      child: groupItemWidget,
                    ),
                    visible: visible,
                  ));
                }
              }

              bool isAllReadOnly = true;
              bool isAllInvisible = true;
              List<T> controllableItems = [];
              items.forEach((element) {
                bool readOnly = element.readOnly;
                bool visible = element.visible;
                if (!readOnly) {
                  isAllReadOnly = false;
                }
                if (visible) {
                  isAllInvisible = false;
                }
                if (!readOnly && visible) {
                  controllableItems.add(element.data);
                }
              });

              Widget? icon;

              if (items.length > 1 &&
                  allowSelectAll &&
                  type != FormeListTileType.Radio) {
                bool selectAll = controllableItems.isNotEmpty &&
                    controllableItems
                        .every((element) => state.value.contains(element));

                if (!isAllInvisible) {
                  void toggleValues() {
                    state.requestFocus();
                    List<T> values = List.of(state.value);
                    if (selectAll) {
                      state.didChange(values
                          .where(
                              (element) => !controllableItems.contains(element))
                          .toList());
                    } else {
                      state.didChange(values
                        ..removeWhere(
                            (element) => controllableItems.contains(element))
                        ..addAll(controllableItems)
                        ..toSet()
                        ..toList());
                    }
                  }

                  icon = InkWell(
                    child: IconButton(
                      icon: (selectAll
                              ? state.model.selectAllIcon
                              : state.model.unselectAllIcon) ??
                          Icon(selectAll
                              ? Icons.switch_right
                              : Icons.switch_left),
                      onPressed:
                          readOnly || isAllReadOnly ? null : toggleValues,
                    ),
                  );
                }
              }

              Widget child = FormeRenderUtils.wrap(
                  state.model.formeWrapRenderData, wrapWidgets);
              if (split == 1) {
                child = FormeRenderUtils.mergeListTileTheme(
                    child, formeListTileRenderData);
              }

              return FormeDecoration(
                formeDecorationFieldRenderData:
                    state.model.formeDecorationFieldRenderData,
                labelText: state.model.labelText,
                helperText: state.model.helperText,
                errorText: state.errorText,
                child: child,
                focusNode: state.focusNode,
                icon: icon,
              );
            });

  @override
  _FormeListTileState<T> createState() => _FormeListTileState();
}

class _FormeListTileState<T>
    extends NonnullValueFieldState<List<T>, FormeListTileModel<T>> {
  @override
  void beforeUpdateModel(
      FormeListTileModel<T> old, FormeListTileModel<T> current) {
    if (current.items != null) {
      List<T> items = List.of(value);
      Iterable<T> datas = current.items!.map((e) => e.data);
      items.removeWhere((element) => !datas.contains(element));
      setValue(items);
    }
  }
}

class FormeListTileModel<T> extends FormeModel {
  final String? labelText;
  final String? helperText;
  final int? split;
  final List<FormeListTileItem<T>>? items;
  final bool? allowSelectAll;
  final Widget? selectAllIcon;
  final Widget? unselectAllIcon;
  final FormeListTileRenderData? formeListTileRenderData;
  final FormeCheckboxRenderData? formeCheckboxRenderData;
  final FormeRadioRenderData? formeRadioRenderData;
  final FormeSwitchRenderData? formeSwitchRenderData;
  final FormeWrapRenderData? formeWrapRenderData;
  final FormeDecorationRenderData? formeDecorationFieldRenderData;

  FormeListTileModel({
    this.labelText,
    this.split,
    this.items,
    this.allowSelectAll,
    this.formeListTileRenderData,
    this.formeCheckboxRenderData,
    this.formeRadioRenderData,
    this.formeSwitchRenderData,
    this.formeWrapRenderData,
    this.formeDecorationFieldRenderData,
    this.helperText,
    this.selectAllIcon,
    this.unselectAllIcon,
  });

  @override
  FormeListTileModel<T> copyWith({
    Optional<String>? labelText,
    Optional<String>? helperText,
    Optional<int>? split,
    List<FormeListTileItem<T>>? items,
    bool? allowSelectAll,
    Optional<Widget>? selectAllIcon,
    Optional<Widget>? unselectAllIcon,
    Optional<FormeListTileRenderData>? formeListTileRenderData,
    Optional<FormeCheckboxRenderData>? formeCheckboxRenderData,
    Optional<FormeRadioRenderData>? formeRadioRenderData,
    Optional<FormeSwitchRenderData>? formeSwitchRenderData,
    Optional<FormeWrapRenderData>? formeWrapRenderData,
    Optional<FormeDecorationRenderData>? formeDecorationFieldRenderData,
  }) {
    return FormeListTileModel<T>(
      labelText: Optional.copyWith(labelText, this.labelText),
      helperText: Optional.copyWith(helperText, this.helperText),
      split: Optional.copyWith(split, this.split),
      items: items ?? this.items,
      allowSelectAll: allowSelectAll ?? allowSelectAll,
      selectAllIcon: Optional.copyWith(selectAllIcon, this.selectAllIcon),
      unselectAllIcon: Optional.copyWith(unselectAllIcon, this.unselectAllIcon),
      formeListTileRenderData: Optional.copyWith(
          formeListTileRenderData, this.formeListTileRenderData),
      formeCheckboxRenderData: Optional.copyWith(
          formeCheckboxRenderData, this.formeCheckboxRenderData),
      formeRadioRenderData:
          Optional.copyWith(formeRadioRenderData, this.formeRadioRenderData),
      formeSwitchRenderData:
          Optional.copyWith(formeSwitchRenderData, this.formeSwitchRenderData),
      formeWrapRenderData:
          Optional.copyWith(formeWrapRenderData, this.formeWrapRenderData),
      formeDecorationFieldRenderData: Optional.copyWith(
          formeDecorationFieldRenderData, this.formeDecorationFieldRenderData),
    );
  }
}
